'''

Copyright (C) 2017-2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from sregistry.auth import read_client_secrets
from sregistry.main import ApiConnection
import json
import os

from .auth import authorize
from .build import build
from .pull import pull
from .push import push
from .delete import delete
from .query import (
    search,
    search_all,
    collection_search,
    label_search,
    container_search,
)

class Client(ApiConnection):

    def __init__(self, secrets=None, base=None, **kwargs):
 
        self.base = base
        super(Client, self).__init__(**kwargs)
        self._update_secrets()
        self._update_headers()

    def _update_base(self):
        if self.base is not None:
            self.base = self.base.strip('/')
            if not self.base.endswith('api'):
                self.base = "%s/api" % self.base


    def _read_response(self,response, field="detail"):
        '''attempt to read the detail provided by the response. If none, 
        default to using the reason'''

        try:
            message = json.loads(response._content.decode('utf-8'))[field]
        except:
            message = response.reason
        return message


    def _add_https(self, q):
        '''for push, pull, and other api interactions, the user can optionally
           define a custom registry. If the registry name doesn't include http
           or https, add it.
 
           Parameters
           ==========
           q: the parsed image query (names), including the original
        '''

        # If image uses http or https, add back
        if not q['registry'].startswith('http'):

            if q['original'].startswith('http:'):
                q['registry'] = 'http://%s' % q['registry']

            elif q['original'].startswith('https:'):
                q['registry'] = 'https://%s' % q['registry']

            # Otherwise, guess from the user's environment
            else:

                prefix = 'https://'

                # The user can set an environment variable to specify nohttps
                nohttps = os.environ.get('SREGISTRY_REGISTRY_NOHTTPS')
                if nohttps is not None:
                    prefix = 'http://'
                q['registry'] = '%s%s' %(prefix, q['registry'])

        return q


    def _update_secrets(self):
        '''update secrets will take a secrets credential file
           either located at .sregistry or the environment variable
           SREGISTRY_CLIENT_SECRETS and update the current client 
           secrets as well as the associated API base.
        '''
        self.secrets = read_client_secrets()
        if self.secrets is not None:
            if "registry" in self.secrets:
                if "base" in self.secrets['registry']:
                    self.base = self.secrets['registry']['base']
                    self._update_base()

    def __str__(self):
        return type(self)



Client.authorize = authorize
Client.build = build
Client.delete = delete
Client.pull = pull
Client.push = push
Client.search = search
Client._search_all = search_all
Client._collection_search = collection_search
Client._container_search = container_search
Client._label_search = label_search



'''

Copyright (C) 2016-2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from sregistry.logger import bot

def main(args, parser, extra):

    from sregistry.main import get_client

    image = args.image
    cli = get_client(image, quiet=args.quiet)
    cli.announce(args.command)

    # If the client doesn't have the command, exit
    if not hasattr(cli, 'delete'):
        msg = "delete is not implemented for %s. Why don't you add it?"
        bot.exit(msg % cli.client_name)

    cli.delete(image=image, force=args.force)



'''

Copyright (C) 2018-2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from sregistry.logger import bot
from sregistry.utils import (
    get_file_hash,
    get_recipe_tag,
    get_tmpdir,
    parse_image_name,
    remove_uri
)

from sregistry.main.google_build.utils import get_build_template

try:
    from urllib.parse import unquote
except:
    from urllib2 import unquote

from glob import glob
import time
import tarfile
import shutil
import json
import os
import re

################################################################################
# Manual Build Logic
# 1. Intended for run on the command line, simply use the client to run sregistry
#    build, provide a recipe and a name, and it will update the terminal with
#    status until finished using the build function here. 
#     - If you provide a GitHub uri in the recipe, instead we run build_repo 
#       and get the recipe file from there.
#     - If headless is False, we assume running from the sregistry client and
#       wait for the build to finish, returning the URL to the user.
#     - If headless is True, we assume the function is being run outside of
#       the client, and build_status and build_finish can be used for updates.
#     - if a webhook is provided, headless must be True, and a response is
#       sent here with curl (under development)

def build(self, name,
                recipe="Singularity",
                context=None, 
                preview=False,
                headless=False,
                working_dir=None,
                webhook=None,
                timeout=10800,
                extra_data=None):
    '''trigger a build on Google Cloud (builder then storage) given a name
       recipe, and Github URI where the recipe can be found. This means
       creating and uploading a build package to use for the build.
    
       Parameters
       ==========
       recipe: the local recipe to build.
       name: should be the complete uri that the user has requested to push.
       context: the dependency files needed for the build. If not defined, only
                the recipe is uploaded.
       preview: if True, preview but don't run the build
       working_dir: The working directory for the build. Defaults to pwd.
       timeout: the number of seconds for the build to timeout. The default 
                is 3 hours, and the maximum is 24 hours. If unset (None)
                it will be 10 minutes.
       webhook: if not None, add a curl POST to finish the build. 
       headless: If true, don't track the build, but submit and provide
                 an endpoint to send a response to.
       extra_data: a dictionary of extra_data to send back to the webhook (they
                   are passed in the environment)

       Environment
       ===========
       SREGISTRY_GOOGLE_BUILD_SINGULARITY_VERSION: the version of Singularity
           to use, defaults to v3.2.1-slim
       SREGISTRY_GOOGLE_BUILD_CLEANUP: after build, delete intermediate 
           dependencies in cloudbuild bucket.

    '''
    bot.debug("BUILD %s" % recipe)

    build_package = [recipe]

    if context:

        # If the user gives a ., include recursive $PWD
        if '.' in context:
            context = glob(os.getcwd() + '/**/*', recursive=True)
        build_package = build_package + context
        
    # We need to get and save relative paths for the config.
    package = create_build_package(build_package, working_dir)

    # Does the package already exist? If the user cached, it might
    destination='source/%s' % os.path.basename(package)
    blob = self._build_bucket.blob(destination)

    # if it doesn't exist, upload it
    if not blob.exists() and preview is False:
        bot.log('Uploading build package!')
        self._upload(source=package, 
                     bucket=self._build_bucket,
                     destination=destination)
    else:
        bot.log('Build package found in %s.' % self._build_bucket.name)

    # This returns a data structure with collection, container, based on uri
    names = parse_image_name(remove_uri(name))

    prefix = "%s/" % names['collection']

    # The name should include the complete uri so it's searchable
    name = os.path.basename(names['uri'])

    # Update the recipe name to use a relative path
    recipe = get_relative_path(recipe, working_dir)

    # Load the build configuration (defaults to local)
    config = self._load_build_config(name=name,
                                     prefix=prefix,
                                     recipe=recipe)
    # Add a webhook, if defined
    if webhook and headless:
        config = add_webhook(config=config, 
                             webhook=webhook, 
                             extra_data=extra_data)
 
    # The source should point to the bucket with the .tar.gz, latest generation
    config["source"]["storageSource"]['bucket'] = self._build_bucket.name
    config["source"]["storageSource"]['object'] = destination

    # If the user wants a timeout
    if timeout is not None:
        config['timeout'] = "%ss" % timeout

    # If not a preview, run the build and return the response
    if not preview:
        if not headless:
            config = self._run_build(config)
        else:
            config = self._submit_build(config)

        # If the user wants to cache cloudbuild files, this will be set
        env = 'SREGISTRY_GOOGLE_BUILD_CACHE'
        if not self._get_and_update_setting(env, self.envars.get(env)):
            if headless is False:
                blob.delete()

    # Clean up either way, return config or response
    shutil.rmtree(os.path.dirname(package))
    return config


################################################################################
# GitHub Build Logic
# 1. Intended for run from a server, we submit a job that uses a Github repo#
#    as a parameter, and if there is a response url provided, the finished
#    build will send a response to that server with curl (recommended).
#    If a private repository is provided, it must be prefixed with a token
    
def build_repo(self, 
               repo,
               recipe,
               commit=None,
               branch=None,
               headless=False,
               preview=False,
               webhook=None,
               timeout=10800,
               token=None,
               extra_data=None):
    '''trigger a build on Google Cloud (builder then storage) given a
       Github repo where a recipe can be found. We assume that github.com (or
       some other Git repo) is provided in the name (repo).
    
       Parameters
       ==========
       repo: the full repository address
       recipe: the local recipe to build.
       headless: if True, return the first response (and don't wait to finish)
       commit: if not None, check out a commit after clone.
       branch: if defined, checkout a branch.
       token: if an authentication token is provided, add to GitHub clone url.
       timeout: the number of seconds for the build to timeout. The default 
                is 3 hours, and the maximum is 24 hours. If unset (None)
                it will be 10 minutes.
       webhook: if not None, add a curl POST to finish the build. 
       extra_data: a dictionary of extra_data to send back to the webhook (they
                   are passed in the environment)
    '''
    bot.debug("BUILD %s" % recipe)

    # This returns a data structure with collection, container, based on uri
    names = parse_image_name(remove_uri(repo))

    # Strip git from the end
    if repo.endswith('git'):
        repo = repo.rstrip('.git')

    # In case they added a tag, strip
    if not repo.startswith('http') and not repo.startswith("git@"):
        repo = "https://%s" % repo

    bot.debug("REPO %s" % repo)

    # First preference to command line, then recipe tag
    tag = get_recipe_tag(recipe) or names.get('tag')

    # Update the tag, if recipe provides one
    names = parse_image_name(remove_uri(repo), tag=tag)

    # <collection>/<image>/<tag>/<commit|branch>/<container>.sif
    prefix = "%s/" % names['collection']

    # The name should include the complete uri so it's searchable
    name = os.path.basename(names['uri'])

    if commit or branch:
        name = "%s@%s" %(name, (commit or branch))

    # Load the build configuration
    config = self._load_build_config(template="singularity-cloudbuild-git.json",
                                     prefix=prefix,
                                     name=name,
                                     recipe=recipe)

    # If we need to checkout a particular commit
    if commit or branch:
        commit = commit or branch
        config['steps'].insert(0, {'args': ['checkout', commit],
                                   'name': 'gcr.io/cloud-builders/git'})

    # If the repo is private, add the token
    clone_url = repo
    if token is not None:
        clone_url = "https://%s@%s" %(token, remove_uri(repo))

    # Add the GitHub repo to the recipe, clone to $PWD (.)
    config['steps'].insert(0, {'args': ['clone', clone_url, "."],
                               'name': 'gcr.io/cloud-builders/git'})

    # If the user wants a timeout
    if timeout is not None:
        config['timeout'] = "%ss" % timeout

    # Add the webhook step, if applicable.
    if webhook and headless:
        config = add_webhook(config=config, 
                             webhook=webhook, 
                             extra_data=extra_data)

    # If not a preview, run the build and return the response
    if not preview:
        if not headless:
            return self._run_build(config)
        return self._submit_build(config)

    # If preview is True, we return the config
    return config

def create_build_package(package_files, working_dir=None):
    '''given a list of files, copy them to a temporary folder,
       compress into a .tar.gz, and rename based on the file hash.
       Return the full path to the .tar.gz in the temporary folder.

       Parameters
       ==========
       package_files: a list of files to include in the tar.gz
       working_dir: If set, the path derived for the recipe and
                    files is relative to this.

    '''
    # Ensure package files all exist
    for package_file in package_files:
        if not os.path.exists(package_file):
            bot.exit('Cannot find %s.' % package_file)

    bot.log('Generating build package for %s files...' % len(package_files))
    build_dir = get_tmpdir(prefix="sregistry-build")
    build_tar = '%s/build.tar.gz' % build_dir
    tar = tarfile.open(build_tar, "w:gz")

    # Create the tar.gz, making sure relative to working_dir
    for package_file in package_files:
 
        # Get a relative path
        relative_path = get_relative_path(package_file, working_dir)
        tar.add(package_file, arcname=relative_path)
    tar.close()

    # Get hash (sha256), and rename file
    sha256 = get_file_hash(build_tar)
    hash_tar =  "%s/%s.tar.gz" %(build_dir, sha256)
    shutil.move(build_tar, hash_tar)
    return hash_tar


def get_relative_path(filename, working_dir=None):
    '''given a filename and a working directory, return
       a relative path for the builder to use. This means
       removing the working directory from the filename

       Parameters
       ==========
       filename: the name of the file to get a path for
       working_dir: if not None, remove from paths.
    '''
    relative_path = filename

    # If a working directory is provided, it must end with / 
    if working_dir is not None:
        if not working_dir.endswith(os.sep):
            working_dir = "%s%s" %(working_dir, os.sep)

    if working_dir is not None:
        relative_path = filename.replace(working_dir, '') 
    return relative_path


def add_webhook(config, webhook, extra_data=None):
    '''add a webhook to a config. We assume that the sha256 is calculated in 
       the present working directory. Optionally, the user can provide
       extra_data to post back with the build_id.
    '''
    data = {'id':'$BUILD_ID'}
    substitutions = {}

    if extra_data is not None:

        # Keep a list of envars to add
        for key, val in extra_data.items(): 
            sub = "_SREGISTRY_%s" % key.upper()
            data[key] = "${%s}" % sub
            substitutions[sub] = val

    config['steps'].append({
        "name": "gcr.io/cloud-builders/curl",
        "args":  ["-d", json.dumps(data), "-X", "POST", webhook]})

    config['substitutions'] = substitutions
    return config


def load_build_config(self, name, recipe, 
                            template="singularity-cloudbuild-local.json", 
                            version="v3.2.1-slim",
                            prefix=""):
    '''load a google compute config, meaning that we start with a template,
       and mimic the following example cloudbuild.yaml:

        steps:
        - name: "singularityware/singularity:${_SINGULARITY_VERSION}"
          args: ['build', 'julia-centos-another.sif', 'julia.def']
        artifacts:
          objects:
            location: 'gs://sregistry-gcloud-build-vanessa'
            paths: ['julia-centos-another.sif']


        Parameters
        ==========
        recipe: the local recipe file for the builder.
        name: the name of the container, based on the uri
        template: the template to use, will be populated based on name
        prefix: a prefix for the storage path, defaults to empty string.
    '''
    version_envar = 'SREGISTRY_GOOGLE_BUILD_SINGULARITY_VERSION'
    version = self._get_and_update_setting(version_envar, version)

    # Get the build template based on its name
    config = get_build_template(template)

    # Name is in format 'dinosaur/container:latest'
    storage_name = '%s%s.sif' %(prefix, name)
    container_name = os.path.basename(storage_name)

    # We need to create the equivalent directory for the image
    folder_name = os.path.dirname(storage_name)

    # The bucket location should have the uri
    bucket_location = "gs://%s/%s/" % (self._bucket_name, folder_name)

    # Last step: tag the blog with the sha256 sum
    bucket_sif = '%s%s' %(bucket_location, container_name)
    metadata_command = "gsutil setmeta -h x-goog-meta-sha256sum:$(cat SHA256SUM) %s" % bucket_sif
    config['steps'].insert(0, {'args': ["-c", metadata_command],
                               'entrypoint': "/bin/bash",
                               'name': 'gcr.io/cloud-builders/gsutil'})

    # Second to last step: manually upload the blob
    config['steps'].insert(0, {'args': ['cp', container_name, bucket_sif ],
                               'name': 'gcr.io/cloud-builders/gsutil'})

    # Calculate a hash (happens after build)
    sha256_command = 'sha256sum %s | cut -c 1-64 > SHA256SUM' % container_name
    config['steps'].insert(0, {'args': ['-c', sha256_command],
                               'entrypoint': '/bin/bash',
                               'name': 'ubuntu'})

    # Insert the build step (second step)
    config['steps'].insert(0, {'args': ['build', container_name, recipe],
                               'name': 'singularityware/singularity:%s' % version})

    return config
                

def run_build(self, config):
    '''run a build, meaning creating a build. Retry if there is failure
    '''

    response = self._submit_build(config)
    status = response['metadata']['build']['status']
    build_id = response['metadata']['build']['id']

    start = time.time()
    while status not in ['COMPLETE', 'FAILURE', 'SUCCESS', 'TIMEOUT']:
        time.sleep(15)
        response = self._build_status(build_id)
        status = response['status']
    
    end = time.time()
    bot.log('Total build time: %s seconds' % (round(end - start, 2)))   
    return self._finish_build(build_id, response=response, config=config)


def submit_build(self, config):
    '''run a build, meaning creating a build. Retry if there is failure
    '''

    project = self._get_project()

    #          prefix,    message, color
    bot.custom('PROJECT', project, "CYAN")
    for i, step in enumerate(config['steps']):
        bot.custom('BUILD %s' % i , step['name'], "CYAN")

    response = self._build_service.projects().builds().create(body=config, 
                                              projectId=project).execute()

    build_id = response['metadata']['build']['id']
    status = response['metadata']['build']['status']
    bot.log("build %s: %s" % (build_id, status))

    return response


def build_status(self, build_id):
    '''get a build status based on a build id. We return the entire response
       object for the client to parse.
    '''
    project = self._get_project()
    response = self._build_service.projects().builds().get(id=build_id, 
                                              projectId=project).execute()


    build_id = response['id']
    status = response['status']
    bot.log("build %s: %s" % (build_id, status))

    return response
   

def finish_build(self, build_id, 
                       config=None, 
                       response=None):

    '''finish a build, meaning if the build was successful, we check the
       user preference to make it private. If it's set, we leave it 
       private. Otherwise, we make it public (default).

       Parameters
       ==========
       build_id: the build id returned from submission to track the build.
       config: optionally provide a config to get metadata from for the blob
       response: if we've already gotten a response status, include here
    '''
    # Get the build status, we will only complete on SUCCESS
    if not response:
        response = self._build_status(build_id)

    # If successful, update blob metadata and visibility
    if response['status'] == 'SUCCESS':

        # Does the user want to keep the container private?
        env = 'SREGISTRY_GOOGLE_STORAGE_PRIVATE'
        blob_location = get_blob_location(response, self._bucket_name)
        blob = self._bucket.get_blob(blob_location)
        
        # Make Public, if desired
        if not self._get_and_update_setting(env, self.envars.get(env)):
            blob.make_public()
            response['public_url'] = unquote(blob.public_url)

        # Add the metadata directly to the object, return updaated response
        response = update_blob_metadata(blob=blob, 
                                        response=response,
                                        config=config, 
                                        bucket=self._bucket)

    return response


def get_blob_location(response, bucket):
    '''return a relative path for a blob based on finding the build step,
       and the container built from it.

       Parameters
       ==========
       response: the response from client._build_status(build_id)
       bucket: the name of the build bucket.

    '''
    # Find the build step, it uses singularityware as a builder
    build_step = [x for x in response['steps'] 
                    if x['name'] == 'gcr.io/cloud-builders/gsutil' and
                    x['args'][0] == 'cp']

    # This is fragile, but we have to get the container name
    if len(build_step) > 0:
        location = build_step[0]['args'][-1]
        return re.sub('(gs://%s|%s)' %(bucket, bucket), '', location).strip('/')
    bot.exit("Cannot find build step with image name.")


def update_blob_metadata(blob, response, bucket, config=None):
    '''a specific function to take a blob, along with a SUCCESS response
       from Google build, the original config, and update the blob 
       metadata with the artifact file name, dependencies, and image hash.
    '''
    metadata = blob.metadata or {}

    metadata.update({'crc32c': blob.crc32c,
                     'media_link': blob.media_link,
                     'self_link': blob.self_link,
                     'md5sum': blob.md5_hash,
                     'size': blob.size,
                     'type': "container"}) # identify that the blob is container

    # If a configuration is provided
    if config:
        metadata['buildCommand'] = ' '.join(config['steps'][0]['args'])
        metadata['builder'] = config['steps'][0]['name']

        # If a source was used, add it.
        if "source" in config:
            metadata['storageSourceBucket'] = config['source']['storageSource']['bucket']
            metadata['storageSourceObject'] = config['source']['storageSource']['object']

    # Update the response, keeping data structure flat (and we don't overwrite)
    response.update(metadata)
    blob.metadata = metadata
    blob._properties['metadata'] = metadata
    blob.patch()
    return response



'''

Copyright (C) 2018-2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from sregistry.logger import bot
import json
import sys
import re

def main(args, parser, extra):
    from sregistry.main import get_client

    if args.name is None:
        msg = "You must add the --name of a container uri to build."
        bot.exit(msg)

    cli = get_client(image=args.name, quiet=args.quiet)
    cli.announce(args.command)

    # If the client doesn't have the command, exit
    if not hasattr(cli, 'build'):
        msg = "build is not implemented for %s. Why don't you add it?"
        bot.exit(msg % cli.client_name)

    # Singularity Registry Server uses build with a recipe
    if cli.client_name == 'google-build':
        response = run_google_build(cli, args)
        
    elif cli.client_name == "google-storage":
        response = run_compute_build(cli, args)

    # Currently allows for google_build
    else:
        bot.warning("No URI specified, assuming Singularity Registry with Builder")
        response = run_registry_build(cli, args, extra)

    # If the client wants to preview, the config is returned
    if args.preview:
        print(json.dumps(response, indent=4, sort_keys=True))


def run_google_build(cli, args):
    '''a helper function to control build for the Google Build Client.
       the user can request a local recipe build, or a build from
       a GitHub repo. In both cases, if no recipe is provided, we default
       to Singularity at the root of the PWD or GitHub repository.
    '''
    if args.name is None:
        bot.exit('Please provide a container identifier with --name')

    # Default to recipe "Singularity" unless other is provided
    recipe = "Singularity"
    if args.commands:
        recipe = args.commands.pop(0)

    # If Github.com is provided in the name, we are doing a GitHub build
    if re.search("github.com|gitlab.com", args.name):
        name = args.name.replace('google-build://', '')
        response = cli.build_repo(repo=name,
                                  recipe=recipe,
                                  preview=args.preview)

    else:        
        response = cli.build(name=args.name,
                             recipe=recipe,
                             context=args.commands,
                             preview=args.preview)

    # Print output to the console
    if not args.preview:
        print_output(response, args.outfile)
    return response


def run_compute_build(cli, args):
    '''a compute based build is the oldest versions of build - here we bring
       up our own instance, and then provide control to it. The helper
       functions below (kill, instances, templates) support this version.
    '''
    # Does the user want to save the image?    
    command = args.commands.pop(0)

    # Option 1: The user wants to kill an instance
    if command == "kill":
        kill(args)    

    # Option 2: Just list running instances
    elif command == "instances":
        instances(args)

    # Option 3: The user wants to list templates
    elif 'template' in command:
        templates(args)

    # Option 4: View a specific or latest log
    elif command == 'logs':
        list_logs(args)

    # Option 3: The user is providing a Github repo!
    recipe = "Singularity"

    if "github" in command:

        # One argument indicates a recipe
        if len(args.commands) == 1:
            recipe = args.commands.pop(0)       

    else:
        # If a command is provided, but not a Github repo
        bot.exit('%s is not a recognized option.' % command)

    # Does the user want to specify a name for the collection?
    name = args.name

    # No image is needed, we are creating in the cloud
    return cli.build(repo=command, 
                     name=name,
                     recipe=recipe,
                     preview=args.preview)


def run_registry_build(cli, args, extra):
    '''a registry build pushes a recipe file to Singularity Registry Server,
       or given that a GitHub Url is provided, we build from there. For more
       regular building, the user is suggested to directly connect the 
       repository to Singularity Registry server. This can serve as a one time
       build.
    '''
    # The uri can also contain github, which indicates a Github build
    if args.name is None:
        bot.exit('Please provide a container identifier with --name')

    recipe = args.commands.pop(0)
    response = cli.build(name=args.name,
                         recipe=recipe,
                         extra=extra)

    # Print output to the console
    if response is not None:
        print_output(response, args.outfile)
    return response


def print_output(response, output_file=None):
    '''print the output to the console for the user. If the user wants the content
       also printed to an output file, do that.

       Parameters
       ==========
       response: the response from the builder, with metadata added
       output_file: if defined, write output also to file

    '''
    # If successful built, show container uri
    if "status" in response:
        if response['status'] == 'SUCCESS':
            bucket = response['artifacts']['objects']['location']
            obj = response['artifacts']['objects']['paths'][0]
            bot.custom("MD5HASH", response['file_hash'], 'CYAN')
            bot.custom("SIZE", response['size'], 'CYAN')
            bot.custom(response['status'], bucket + obj , 'CYAN')
        else:
            bot.custom(response['status'], 'see logs for details', 'CYAN')

        # Show the logs no matter what
        bot.custom("LOGS", response['logUrl'], 'CYAN')

        # Did the user make the container public?
        if "public_url" in response:
            bot.custom('URL', response['public_url'], 'CYAN')
  
    # Does the user also need writing to an output file?
    if output_file is not None:    
        with open(output_file, 'w') as filey:
            if response['status'] == 'SUCCESS':
                filey.writelines('MD5HASH %s\n' % response['file_hash'])    
                filey.writelines('SIZE %s\n' % response['size'])    
            filey.writelines('%s %s%s\n' % (response['status'], bucket, obj))
            filey.writelines('LOGS %s\n' % response['logUrl'])
            if "public_url" in response:
                filey.writelines('URL %s\n' % response['public_url'])


def kill(args):
    '''kill is a helper function to call the "kill" function of the client,
       meaning we bring down an instance.
    '''
    from sregistry.main import get_client
    cli = get_client(quiet=args.quiet)
    if len(args.commands) > 0:
        for name in args.commands:
            cli.destroy(name)
    sys.exit(0)

def instances(args):
    '''list running instances for a user, including all builders and report
       instance names and statuses.
    '''
    from sregistry.main import get_client
    cli = get_client(quiet=args.quiet)
    cli.list_builders()
    sys.exit(0)

def templates(args, template_name=None):
    '''list a specific template (if a name is provided) or all templates
       available.

       Parameters
       ==========
       args: the argparse object to look for a template name
       template_name: if not set, show all

    '''
    from sregistry.main import get_client

    # We don't need storage/compute connections
    cli = get_client(init=False, quiet=args.quiet)

    if len(args.commands) > 0:
        template_name = args.commands.pop(0)
    cli.list_templates(template_name)
    sys.exit(0)


def list_logs(args, container_name=None):
    '''list a specific log for a builder, or the latest log if none provided

       Parameters
       ==========
       args: the argparse object to look for a container name
       container_name: a default container name set to be None (show latest log)

    '''
    from sregistry.main import get_client
    cli = get_client(quiet=args.quiet)
    if args.commands:
        container_name = args.commands.pop(0)
    cli.logs(container_name)
    sys.exit(0)



'''

build.py: push a Singularity recipe to Singularity Registry Server to build

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from sregistry.logger import bot, ProgressBar
from sregistry.utils import (
    parse_image_name,
    remove_uri
)

from requests_toolbelt import (
    MultipartEncoder,
    MultipartEncoderMonitor
)

import requests
import os


def build(self, recipe, name, extra=None):
    '''push a recipe file to Singularity Registry server for a Google
       Cloud (or similar) build
    '''
    path = os.path.abspath(recipe)
    bot.debug("BUILD %s" % path)

    if extra is None:
        extra = {}

    if not os.path.exists(path):
        bot.exit('%s does not exist.' % path)

    if not os.path.isfile(path):
        bot.exit("Build takes a Singularity recipe file")

    # Interaction with a registry requires secrets
    self.require_secrets()

    valid = ['google_build']

    # Only one valid type
    if "google_build" not in extra:
        bot.exit('Please include --builder google_build as the last extra arugment for Google Cloud Build')

    builder_type = None
    for builder_type in extra:
        if builder_type in valid:
            break

    # Must have valid builder type
    if builder_type is None:
        bot.exit("Invalid builder type.")

    # Extract the metadata
    names = parse_image_name(remove_uri(name))

# COLLECTION ###################################################################

    # If the registry is provided in the uri, use it
    if names['registry'] is None:
        names['registry'] = self.base

    # If the base doesn't start with http or https, add it
    names = self._add_https(names)

    # Recipe filename to upload to
    upload_to = "Singularity.%s-%s-%s" % (names['collection'],
                                          names['image'],
                                          names['tag'])

    # Prepare build request
    url = '%s/%s/build/' %(names['registry'].replace('/api', ''), builder_type)
    SREGISTRY_EVENT = self.authorize(request_type="build", names=names)
    headers = { 'Authorization': SREGISTRY_EVENT }

    bot.debug('Setting build URL to {0}'.format(url))

    # Fields for build endpoint
    fields={'SREGISTRY_EVENT': SREGISTRY_EVENT,
            'name': names['image'],
            'collection': names['collection'],
            'tag': names['tag'],
            'datafile': (upload_to, open(path, 'rb'), 'text/plain')}

    encoder = MultipartEncoder(fields=fields)
    progress_callback = create_callback(encoder, self.quiet)
    monitor = MultipartEncoderMonitor(encoder, progress_callback)
    headers = {'Content-Type': monitor.content_type,
               'Authorization': SREGISTRY_EVENT }

    try:
        r = requests.post(url, data=monitor, headers=headers)
        r.raise_for_status()
        print('\n[Return status {0} Created]'.format(r.status_code))
    except requests.HTTPError as e:
        print('\nRecipe upload failed: {0}.'.format(e))
    except KeyboardInterrupt:
        print('\nRecipe upload cancelled.')


def create_callback(encoder, quiet=False):
    encoder_len = encoder.len / (1024*1024.0)

    bar = ProgressBar(expected_size=encoder_len,
                      filled_char='=',
                      hide=quiet)

    def callback(monitor):
        bar.show(monitor.bytes_read / (1024*1024.0))
    return callback



#!/usr/bin/python

# Copyright (C) 2017-2019 Vanessa Sochat.

# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import shutil
import pytest


def test_write_read_files(tmp_path):
    '''test_write_read_files will test the functions write_file and read_file
    '''
    print("Testing utils.write_file...")
    from sregistry.utils import write_file
    tmpfile = str(tmp_path / 'written_file.txt')
    assert not os.path.exists(tmpfile)
    write_file(tmpfile, "hello!")
    assert os.path.exists(tmpfile)

    print("Testing utils.read_file...")
    from sregistry.utils import read_file
    content = read_file(tmpfile)[0]
    assert content == "hello!"

def test_write_bad_json(tmp_path):
    from sregistry.utils import write_json
    bad_json = {"Wakkawakkawakka'}": [{True}, "2", 3]}
    tmpfile = str(tmp_path / 'json_file.txt')
    assert not os.path.exists(tmpfile)   
    with pytest.raises(TypeError):
        write_json(bad_json, tmpfile)

def test_write_json(tmp_path):
    import json
    from sregistry.utils import write_json, read_json
    good_json = {"Wakkawakkawakka": [True, "2", 3]}
    tmpfile = str(tmp_path / 'good_json_file.txt')
    assert not os.path.exists(tmpfile)
    write_json(good_json, tmpfile)
    with open(tmpfile, 'r') as f:
        content = json.loads(f.read())
    assert isinstance(content, dict)
    assert "Wakkawakkawakka" in content
    content = read_json(tmpfile)
    assert "Wakkawakkawakka" in content

def test_check_install():
    '''check install is used to check if a particular software is installed.
    If no command is provided, singularity is assumed to be the test case'''
    print("Testing utils.check_install")
    from sregistry.utils import check_install
    is_installed = check_install('echo')
    assert is_installed
    is_not_installed = check_install('fakesoftwarename')
    assert not is_not_installed

def test_get_installdir():
    '''get install directory should return the base of where sregistry
       is installed
    '''
    print("Testing utils.get_installdir")
    from sregistry.utils import get_installdir
    whereami = get_installdir()
    print(whereami)
    assert whereami.endswith('sregistry')


def test_remove_uri():
    print("Testing utils.remove_uri")
    from sregistry.utils import remove_uri
    assert remove_uri('docker://ubuntu') == 'ubuntu'
    assert remove_uri('shub://vanessa/singularity-images') == 'vanessa/singularity-images'
    assert remove_uri('vanessa/singularity-images') == 'vanessa/singularity-images'


def ubuntu_check(names, original):
    assert names['image'] == 'ubuntu'
    assert names['collection'] == 'library'
    assert names['registry'] is None
    assert names['version'] is None
    assert names['tag'] == 'latest'
    assert names['url'] == "library/ubuntu"
    assert names['uri'] == 'library/ubuntu:latest'
    assert names['original'] == original
    assert names['storage'] == 'library/ubuntu:latest.sif'

def test_parse_image_name():
    print("Testing utils.parse_image_name")
    from sregistry.utils import parse_image_name        
    names = parse_image_name('ubuntu')
    ubuntu_check(names, 'ubuntu')
    names = parse_image_name('ubuntu:latest')
    ubuntu_check(names, 'ubuntu:latest')
    names = parse_image_name('library/ubuntu:latest')
    ubuntu_check(names, 'library/ubuntu:latest')

    # Version should include same with version
    names = parse_image_name('library/ubuntu:latest@version')
    assert names['version'] == 'version'
    assert names['uri'] == "library/ubuntu:latest@version"
    assert names['image'] == 'ubuntu'
    assert names['collection'] == 'library'
    assert names['registry'] is None
    assert names['tag'] == 'latest'
    assert names['url'] == "library/ubuntu"
    assert names['uri'] == 'library/ubuntu:latest@version'
    assert names['original'] == 'library/ubuntu:latest@version'
    assert names['storage'] == 'library/ubuntu:latest@version.sif'


def test_recipe_tag():
    print("Testing utils.get_recipe_tag")
    from sregistry.utils import get_recipe_tag
    assert get_recipe_tag("Singularity") == "latest"
    assert get_recipe_tag("/path/to/Singularity") == "latest"
    assert get_recipe_tag("/path/to/Singularity.tag") == "tag"
    assert get_recipe_tag("/path/to/Dockerfile") is None


def test_run_command():
    print("Testing utils.run_command")
    from sregistry.utils import run_command
    result = run_command(["echo", "hello"])
    assert result['message'] == "hello\n"
    assert result['return_code'] == 0

def test_get_thumbnail():
    print("Testing utils.get_thumbnail")
    from sregistry.utils import get_thumbnail
    assert os.path.exists(get_thumbnail())
    assert get_thumbnail().endswith('robot.png')

def test_which():
    print("Testing utils.which")
    from sregistry.utils import which
    result = which('echo')
    assert result['message'].endswith('echo')
    assert result['return_code'] == 0
    result = which('singularityaaaa')
    assert result['message'] == ''
    assert result['return_code'] == 1


def test_get_file_hash():
    print("Testing utils.get_file_hash")
    from sregistry.utils import get_file_hash
    here = os.path.dirname(os.path.abspath(__file__))
    testdata = os.path.join(here, 'testdata', 'hashtest.txt')
    assert get_file_hash(testdata) == '6bb92117bded3da774363713657a629a9f38eac2e57cd47e1dcda21d3445c67d'
    assert get_file_hash(testdata, 'md5') == 'e5d376ca96081dd561ff303c3a631fd5'


def test_get_uri():
    print("Testing utils.get_uri")
    from sregistry.utils import get_uri
    assert get_uri('docker://ubuntu') == 'docker'
    assert 'hub' in get_uri('shub://vanessa/singularity-images')
    assert get_uri('notvalid://') is None
    assert get_uri('nowvalid://', validate=False) == 'nowvalid'


def test_get_userhome():
    print("Testing utils.get_userhome")
    from sregistry.utils import get_userhome
    home = get_userhome()
    assert home in os.environ.get('HOME')

def test_read_tar(tmp_path):
    print("Testing utils.extract_tar")
    from sregistry.utils import extract_tar
    here = os.path.dirname(os.path.abspath(__file__))
    testdata = os.path.join(here, 'testdata', 'hashtest.tar.gz')
    dirname = str(tmp_path / "destination")
    os.mkdir(dirname)
    extract_tar(testdata, dirname)
    assert os.path.exists(dirname)
    assert "hashtest.txt" in [os.path.basename(x) for x in os.listdir(dirname)]

def test_copyfile(tmp_path):
    print("Testing utils.copyfile")
    from sregistry.utils import copyfile, write_file
    original = str(tmp_path / 'location1.txt')
    dest = str(tmp_path / 'location2.txt')
    print(original)
    print(dest)
    write_file(original, "CONTENT IN FILE")
    copyfile(original, dest)
    assert os.path.exists(original)
    assert os.path.exists(dest)

def test_get_tmpdir_tmpfile():
    print("Testing utils.get_tmpdir, get_tmpfile")
    from sregistry.utils import get_tmpdir, get_tmpfile
    tmpdir = get_tmpdir()
    assert os.path.exists(tmpdir)
    assert os.path.basename(tmpdir).startswith('sregistry')
    shutil.rmtree(tmpdir)
    tmpdir = get_tmpdir(prefix='name')
    assert os.path.basename(tmpdir).startswith('name')
    shutil.rmtree(tmpdir)
    tmpfile = get_tmpfile()
    assert 'sregistry' in tmpfile
    os.remove(tmpfile)
    tmpfile = get_tmpfile(prefix="pancakes")
    assert 'pancakes' in tmpfile
    os.remove(tmpfile)

def test_mkdir_p(tmp_path):
    print("Testing utils.mkdir_p")
    from sregistry.utils import mkdir_p
    dirname = str(tmp_path / "input")
    result = os.path.join(dirname, "level1", "level2", "level3")
    mkdir_p(result)
    assert os.path.exists(result)   

def test_print_json():
    print("Testing utils.print_json")
    from sregistry.utils import print_json
    result = print_json({1:1})
    assert result == '{\n    "1": 1\n}'
